( Mandelbrot Plots, 15jun07cht )
( A F# Tutorial for SVFIG, June 23, 2007 )
( Make high res photos, 08feb10cht )
( Add color controls, 23feb10cht )
( Use SetBitmapBits to display image, 25mar10cht )

\ Mandelbrot plots allows one to learn:
\       Mandelbrot fractals
\       Integer numerical computation for floating point numbers
\       486 assmbly in F#
\       Plotting color images in Windows
\       Running Floating Point Unit FPU in 486 assembly
\ ASMTEST.F in the F# system has many examples of assembly programming.
\ Please use these examples as templates when you start writing
\ FORTH words in 486 assembler.
\ You probably need also an Intel reference manual to learn what each
\ assembly code does.  I have a copy of Intel486 Processor Family
\ Programmer's Reference Manual.  It is enormously helpful, especially
\ on the FPU machine code.

\ 1. Mandelbrot Plot
\ Mandelbrot plots are based on the iterative equations on a complex number
\ (x,y), starting at (x0,y0):
\         (x[n+1],y[n+1]) = (x0,y0) + (x[n],y[n])**2
\ The real part is:
\         x[n+1] = x0 + x[n]**2 - y[n]**2
\ The imaginary part is:
\         y[n+1] = y0 + 2x[n]*y[n]
\ When (x[n],y[n]) is outside of the circle of radius 2.0, it diverges
\ quickly.
\ Each point (x,y) has a unique n associated with it.  Assign a color
\ to this point and plot all the points in color.

\ 2. Scaled Integer Numbers
\ Numbers are represented in a scaled integer format.  Bit31 is a sign
\ bit and bits30-28 are the integer bits.  The decimal point is between
\ bit28 and bit27.  Therefore, 1.0 is 10000000H and -1.0 is F0000000H..
\ 2.0 is 200000000H and -2.0 is E0000000H.

\ The product of two integers must be scaled to the left by 4 bits so
\ that the decimal point is realigned.  Since the Mandelbrot plot
\ diverges when (x*x+y*y)>4.0, a mask C0000000H can be used to terminate
\ the iterative computation of a point.

\ 3. Constants and Variables

decimal

: break cr .s key 27 = abort" done" ;

768 constant width
512 constant height
create palette 1024 cells allot
create color_table 1024 cells allot

variable s
variable x
variable y
variable dx
variable left
variable right
variable top
variable bottom
variable xx
variable yy
variable iteration
variable max_iteration
127 max_iteration !
variable color
16 color !
variable base_color
0 base_color !


\ 4. Color Palette
\ PALETTE is an 1024 entry palette table
\ colors are code in 24 bits of a 32 bit word.  Code format is
\ 00000000rrrrrrrrggggggggbbbbbbbbB
\ Each primary color of red, green and blue has an 8 bit value.
\ FILL_PALETTE
\ Palette table is initialized with a rainbow pattern, with the
\ folowing colors at 128th color boundaries:
\ Black, red, orange, yellow, green, cyne, blue. purple, white
\ FILL_COLOR_TABLE
\ Depending on the value in COLOR, that many colors are selected
\ from the 1024 entry Palette table, and fill the COLOR-TABLE.
\ GET_COLOR_1 converts a Mandelbrot iterative number n to a color code
\ in the PALETTE_TABLE.
\ BASE_COLOR has the initial color to pick 128 colors filling the 
\ color table.
\ COLOR and BASE_COLOR used together allow you to change the color
\ appearance of your Mandelbrot plots.

: fill_palette
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap $10000 + swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $2 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_palette_yellow
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap $10000 + swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_palette_violet
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap $100 + swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_palette_red
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap $10100 + swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_palette_green
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap $10001 + swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_palette_blue
   0 palette
   255 for
      over $FFFFFF xor over !
      cell+ swap $101 + swap
   next
   swap drop $FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   swap drop $FEFE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 - swap
   next
   swap drop $FE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $20000 + swap
   next
   swap drop $FEFE00 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 - swap
   next
   swap drop $FE0000 swap
   127 for
      over $FFFFFF xor over !
      cell+ swap 2 + swap
   next
   swap drop $FE00FE swap
   127 for
      over $FFFFFF xor over !
      cell+ swap $200 + swap
   next
   2drop ;

: fill_color_table
   base_color @ color_table
   1023 for
      over cells $FFF and palette + @
      over !
      cell+ swap
      1024 color @ / + swap
   next
   2drop ;

fill_palette

: colors ( n -- )
   color !
   fill_color_table
   ;
16 colors

: more_colors
   color @ 2* dup 1023 >
   if drop
   else colors
   then
   ;
: less_colors
   color @ 2/ dup 1 >
   if colors 
   else drop
   then
   ;
: higher_colors
   base_color @
   1024 color @ / + $FFF and
   base_color !
   fill_color_table
   ;
: lower_colors
   base_color @
   1024 color @ / - $FFF and
   base_color !
   fill_color_table
   ;

newcanvas

: get_color_1 ( code -- color )
   $3FF and cells color_table + @ ;


\ 5.  Mandelbrot Using Floating Point Unit

\ Use the Floating Point Unit in Intel processor to do computation.
\ It allows numbers to be represented in 80 bit floating point
\ format, and will increase the depths of the Mandelbrot plots.
\ FPU multiplications are much faster than integer multiplications.
\ When all the computations are done inside FPU, a 2x increase in
\ speed is expected..
\ There are two issues: one is the computation inside FPU, and the
\ other is the initiation of FPU with the Mandelbrot coordinates.
\ The first experiments are converting the existing Mandelbrot
\ program to make use of FPU.
\ The 486 assembler in F# can assemble all the FPU instructions.
\ FPU is a classic stack-based computing device.  All operations
\ are focused on the top register on the stack.

variable integer
variable four
variable minus1
variable minus3
variable two
variable divisor
$100 divisor !
0 integer !
2 two !
-1 minus1 !
-3 minus3 !
4 four !

\ Here I am using double integers to represent floating point
\ coordinates and increments.  1.0 is 1000,0000,0000,0000H.
\ 2.0 is 2000,0000,0000,0000. -1.0 is F000,0000,0000,0000H,
\ and -2.0 is E000,0000,0000,0000.

\ FLOAT_X, FLOAT_Y, FLOAT_X0, FLOAT_Y0 are used to store floating
\   point numbers representing Mandelbrot coordinates
\ FLOAT_D is the increment on coordinates between neighboring pixels.
create float_x 10 allot
create float_y 10 allot
create float_x0 10 allot
create float_y0 10 allot
create float_d 10 allot
float_x 10 0 fill
float_y 10 0 fill
float_x0 10 0 fill
float_y0 10 0 fill
float_d 10 0 fill


\ D@, D!, DLSHIFT, DRSHIFT compliment D+ and DNEGATE for
\ 64-bit integer math operations.
\ FLOAT_D*2 and FLOAT_D/2 scales FLOAT_D for zoom-out and zoom-in.
\ DUM* multiplies a double integer and a single interger to get
\   a double integer product.  It is used to calculate the
\   coordinates at the top-left corner from the center coordinates
\   to initiate a Mandelbrot plot.
\ INCREMENT increments the selected coorindates by value in FLOAT_D.
\ XY does Mandelbrot iteration on (FLOAT_X,FLOAT_Y)
\ XX does a Mandelbrot plot.  Center is ar (FLOAT_X0, FLOAT_Y0) and
\   coordinate increment is in FLOAT_D.
\ XXX does an initial Mandelbrot plot.
\ RRR, LLL, UUU, DDD panning Mandelbrot plot right, left, up and down.
\ +++ zooms in on a Mandelbrot plot
\ --- zooms out on a Mandelbrot plot
\ scene_1, scene_2, scene_3, scene_4 various points of interests.
\ MANDELBROT  keyboard interface to explore Mandelbrot plots.
\       key     function
\       = (+)   zoom in
\       -       zoom out
\       u       pan up
\       d       pan down
\       r       pan right
\       l       pan left
\	0	intial plot
\       1       scene_1
\       2       scene_2
\       3       scene_3
\       4       scene_4
\	]	decrease base_color
\	[	increase base_color
\	}	decrease color
\	{	increase color
\       ESC     quit

( use double integers for delta )
( 1.0 = $1000,0000,0000,0000 )
variable justification
$-3c justification !

code D!
        mov     eax, [ebp]
        mov     ecx, 4 [ebp]
        mov     [ebx], ecx
        mov     4 [ebx], eax
        mov     ebx, 8 [ebp]
        lea     ebp, $0c [ebp]
        ret
        end-code

code D@
        lea     ebp, -4 [ebp]
        mov     eax, [ebx]
        mov     [ebp], eax
        mov     ebx, 4 [ebx]
        ret
        end-code

code DLSHIFT ( d -- 2d )
        mov     eax, [ebp]
        shld    eax, ebx, # 1             ( normalize )
        shl     eax, # 1
        mov     [ebp], eax
        ret
        end-code

code DRSHIFT ( d -- 2d )
        mov     eax, [ebp]
        shrd    ebx, eax, # 1             ( normalize )
        shr     ebx, # 1
        mov     [ebp], eax
        ret
        end-code

code DLSHIFT_ ( d -- 2d )
        mov     eax, [ebp]
        shl     eax, # 1             ( normalize )
        rcl     ebx, # 1
        mov     [ebp], eax
        ret
        end-code

code DRSHIFT_ ( d -- 2d )
        mov     eax, [ebp]
        shr     ebx, # 1             ( normalize )
        rcr     eax, # 1
        mov     [ebp], eax
        ret
        end-code

: float_d*2
        float_d D@
        dup $10000000 <
        if DLSHIFT float_d D!
        else 2drop
        then
        ;

: float_d/2
        float_d D@ 2dup
        or
        if DRSHIFT float_d D!
        else 2drop
        then
        ;

code xy ( -- n )
        lea     ebp, -4 [ebp]
        mov     [ebp], ebx
        finit
        fild    dword justification
        fild    qword float_x
        fscale
        fstp    st(1)
        fild    dword justification
        fild    qword float_y
        fscale
        fstp    st(1)                  ( x0 y0 )
        mov     ecx, max_iteration
        mov     ebx, ecx
        fld     st(1)                   ( x0 y0 x )
        fld     st(1)                   ( x0 y0 x y )
@@1:    fld     st(1)                   ( x0 y0 x y x )
        fmul    st, st                  ( x0 y0 x y x*x )
        fld     st(1)                   ( x0 y0 x y x*x y )
        fmul    st, st                  ( x0 y0 x y x*x y*y )
        fld     st(1)                   ( x0 y0 x y x*x y*y x*x )
        fadd    st, st(1)               ( x0 y0 x y x*x y*y x*x+y*y )
        ficomp  dword four              ( x0 y0 x y x*x y*y )
        fstsw   ax
        sahf
        jnb     @@2                     ( x*x+y*y=<4 ? )
        fsubp   st(1), st               ( x0 y0 x y x*x-y*y )
        fld     st(2)                   ( x0 y0 x y x*x-y*y x )
        fmul    st, st(2)               ( x0 y0 x y x*x-y*y xy )
        fadd    st, st                  ( x0 y0 x y x*x-y*y 2xy )
        fadd    st, st(4)               ( x0 y0 x y x*x-y*y y0+2xy )
        fstp    st(2)                   ( x0 y0 x y0+2xy x*x-y*y  )
        fadd    st, st(4)               ( x0 y0 x y0+2xy x0+x*x-y*y )
        fstp    st(2)                   ( x0 y0 x0+x*x-y*y y0+2xy )
        loop    @@1
@@2:    sub     ebx, ecx
        ret
        end-code

: dum* ( d u -- d*u )
        2dup um* drop >r
        nip um* r> + ;
code dum*_
        mov     eax, 4 [ebp]
        mul     eax, ebx
        mov     ecx, edx
        mov     4 [ebp], eax
        mov     eax, [ebp]
        mul     eax, ebx
        mov     ebx, ecx
        add     ebx, edx
        add     ebp, # 4
        ret
        end-code
: increment ( a -- )
        dup >r D@
        float_d D@ D+
        r> D! ;
code increment_
        mov     eax, [ebx]
        add     eax, float_d
        mov     [ebx], eax
        mov     eax, 4 [ebx]
        adc     eax, float_d cell+
        mov     4 [ebx], eax
        mov     ebx, [ebp]
        add     ebp, 4 #
        ret
        end-code

: IMAGE ( bitmap image array in memory)
	PAD ALIGNED ;

\ Use SetBitmapBits GDI call to copy image in memory to hBitmap.
\ hBitmap is the bitmap handle of Canvas window.
\ Using SetBitmapBits instead of SetPixel doubles plotting speed.

: xx
  float_y0 D@ float_d D@ height 2/ dum* dnegate D+ float_y D!
  IMAGE  ( image in memory )
  height for aft
     float_x0 D@ float_d D@ width 2/ dum* dnegate D+ float_x D!
     width for aft
        xy \ break
        get_color_1 OVER ! cell+ ( write to IMAGE )
        float_x increment
     then next
     float_y increment
  then next
  drop
  hBitmap @ $180000 IMAGE SetBitmapBits DROP ( copy IMAGE to Canvas )
  update 
  ;

: xxx ( initial Mandelbrot plot )
   0 $200000 float_d D!
   0 0 float_y0 D!
   0 0 float_x0 D! xx ;

: range? ( x y -- f )
   abs dup um* nip swap abs dup um* nip + $F8000000 and ;

: rrr ( shift plot right by 48 pixels )
   float_x0 cell+ @ float_d cell+ @ $30 * -
   float_y0 cell+ @ range?
   if
   else float_x0 D@ float_d d@ $30 dum* dnegate D+ float_x0 D! xx
   then
   ;

: lll ( shift plot left by 48 pixels )
   float_x0 cell+ @ float_d cell+ @ $30 * +
   float_y0 cell+ @ range?
   if
   else float_x0 D@ float_d d@ $30 dum* D+ float_x0 D! xx
   then
   ;

: uuu ( shift plot up by 48 pixels )
   float_y0 cell+ @ float_d cell+ @ $30 * +
   float_x0 cell+ @ range?
   if
   else float_y0 D@ float_d d@ $30 dum* D+ float_y0 D! xx
   then
   ;

: ddd ( shift plot down by 48 pixels )
   float_y0 cell+ @ float_d cell+ @ $30 * -
   float_x0 cell+ @ range?
   if
   else float_y0 D@ float_d d@ $30 dum* dnegate D+ float_y0 D! xx
   then
   ;

: ---	( shrink plot by a factor of 2 )
        float_d D@
        dup $10000000 <
        if DLSHIFT float_d D! xx
        else 2drop
        then
        ;

: +++	( expand plot by a factor of 2 )
        float_d D@ DRSHIFT 2dup
        or
        if float_d D! xx
        else 2drop
        then
        ;

: scene_1 0 $F7000000 float_x0 d! 0 0 float_y0 d! 0 $200000 float_d d! xx ;
: scene_2 0 $F7000000 float_x0 d! 0 0 float_y0 d! 0 $100000 float_d d! xx ;
: scene_3 0 $e5000000 float_x0 d! 0 0 float_y0 d! 0 1 float_d d! xx ;
: scene_4 0 $e0000000 float_x0 d! 0 0 float_y0 d! 1 0 float_d d! xx ;
: red fill_palette_red $400 colors xx ;
: green fill_palette_green $400 colors xx ;
: blue fill_palette_blue $400 colors xx ;
: cyan fill_palette $400 colors xx ;
: yellow fill_palette_yellow $400 colors xx ;
: violet fill_palette_violet $400 colors xx ;

HEX
: mandelbrot ( start Mandelbrot plot )
   xxx
   begin
        key
        dup 75 = if uuu then
        dup 64 = if ddd then
        dup 6c = if lll then
        dup 72 = if rrr then
        dup 3d = if +++ then
        dup 2d = if --- then
        dup 30 = if xxx 0 base_color ! 10 colors then
        dup 31 = if scene_1 then
        dup 32 = if scene_2 then
        dup 33 = if scene_3 then
        dup 34 = if scene_4 then
        dup 6E = if violet then
        dup 62 = if blue then
        dup 76 = if cyan then
        dup 63 = if green then
        dup 78 = if yellow then
        dup 7A = if red then
        dup 5B = if lower_colors xx then
        dup 5D = if higher_colors xx then
        dup 7B = if less_colors xx then
        dup 7D = if more_colors xx then
        1b =
   until
   ;

\ I was expecting that this Mandelbrot plot could zoom 2**60 times.
\ The tests show that at 2**40, the plots fail to show more details
\ and the fine structures round off.  This is because FPU truncates
\ the product of two floating numbers.  Bits below 40 significant bits
\ do not have any effects on the product.  The product rounds off
\ with 80 significant bits.

\ Mouse interface 24feb10cht
\ The mouse is more intuitive than keyboard interface.
\ I like to have these 3 mouse function:
\ 	Drag a plot around by holding down left button and move mouse
\	Expand the plot around cursor by click right mouse
\	Shrink the plot around cursor by double click left mouse
\ However, double clicking does not work, and I have to use the
\ control key.  To shrink a plot, hold down control key.  Hold down
\ left mouse button and move mouse slight.
\ See canvasE.f for mouse interface code.

: dm* ( d n -- d*n )
	dup 0<
	if negate dum* dnegate
	else dum*
	then
	;

: LButtonDown 
	MOUSEPOS.X mouse_x ! MOUSEPOS.Y mouse_y ! ;
: MouseMove 
	float_d D@ mouse_x @ MOUSEPOS.X - dm* 
	float_x0 D@ D+ float_x0 D!
	float_d D@ mouse_y @ MOUSEPOS.Y - dm* 
	float_y0 D@ D+ float_y0 D!
	xx ;
: LButtonDoubleClick ( expand )
	float_d D@ MOUSEPOS.X $180 - 2/ dm* 
	float_x0 D@ D+ float_x0 D!
	float_d D@ MOUSEPOS.Y $100 - 2/ dm* 
	float_y0 D@ D+ float_y0 D!
	+++ ;
: RButtonDoubleClick ( shrink )
	float_d D@ MOUSEPOS.X $180 - dm* dnegate
	float_x0 D@ D+ float_x0 D!
	float_d D@ MOUSEPOS.Y $100 - dm* dnegate
	float_y0 D@ D+ float_y0 D!
	--- ;
' MouseMove (MOUSEMOVE) !
' LButtonDown (LBUTTONDOWN) !
' LButtonDown (LBUTTONUP) !
' RButtonDoubleClick (RBUTTONDBLCLK) !
' LButtonDoubleClick (RBUTTONDOWN) !

MANDELBROT



	